from pydantic import BaseModel
from artefacts.mixins import ContextReader


class BaseNode(ContextReader, BaseModel):

    unique_id: str

    class Config:
        extra = "allow"
        arbitrary_types_allowed = True

    def __hash__(self):
        return hash(self.unique_id)

    def __str__(self):
        return self.unique_id

    def __repr__(self):
        return f"<Node {self.unique_id}>"


class BaseNodeReference(ContextReader, str):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not isinstance(v, str):
            raise TypeError("NodeReferences must be strings")
        else:
            return cls(v)

    def __str__(self):
        return self.unique_id

    def __repr__(self):
        return f"<Node {self.unique_id}>"



from typing import Union

from artefacts.mixins import ContextReader
from artefacts.catalog import Catalog
from artefacts.manifest import Manifest
from artefacts.run_results import RunResults
from artefacts.sources import Sources


class Project(ContextReader):
    """
    Examples:

        >>> from artefacts import Project
        >>> project = Project(target='./target')

    Attributes:

        manifest (Manifest):
            The project's compiled [`Manifest`][artefacts.manifest.Manifest], which
            includes information about models, tests, sources, seeds, and other
            data about project.

        catalog (Catalog):
            The project's compiled [`Catalog`][artefacts.catalog.Catalog], which
            contains information about the model's actual columns as defined in your
            data warehouse's information_schema tables. Requires that `dbt docs
            generate` was ran before initializing the object, ie before using
            `Project()`.

        run_results (RunResults):
            The project' compiled [`RunResults`][artefacts.run_results.RunResults]
            artifact, which contains details of the results from any `build`, `run`,
            `test` commands.

        sources (Sources):
            The projects compiled [`Sources`][artefacts.sources.Sources] artifact,
            which contains details about the freshness of the project's defined
            sources, as generated by `dbt source snapshot-freshness`.

        models (list):
            A list of the project's compiled models.

        tests (list):
            A list of the project's compiled tests.

        seeds (list):
            A list of the project's compiled seeds.
    """

    manifest: Manifest
    catalog: Union[Catalog, None]
    run_results: Union[RunResults, None]
    sources: Union[Sources, None]

    def __init__(self, target="./target"):
        self.target = target
        self.context.load("manifest", target=self.target, allow_missing=False)
        self.context.parse("manifest", parser=Manifest, allow_missing=False)
        self.context.load("catalog", target=self.target, allow_missing=True)
        self.context.parse("catalog", parser=Catalog, allow_missing=True)
        self.context.load("run_results", target=self.target, allow_missing=True)
        self.context.parse("run_results", parser=RunResults, allow_missing=True)
        self.context.load("sources", target=self.target, allow_missing=True)
        self.context.parse("sources", parser=Sources, allow_missing=True)

    @property
    def models(self):
        return [v for k, v in self.manifest.nodes.items() if v.resource_type == "model"]

    @property
    def tests(self):
        return [v for k, v in self.manifest.nodes.items() if v.resource_type == "test"]

    @property
    def seeds(self):
        return [v for k, v in self.manifest.nodes.items() if v.resource_type == "seed"]

    def get_model(self, name: str):
        """
        Get a model by name.

        Examples:

            >>> from artefacts import Project
            >>> project = Project(target='./target')
            >>> model = project.get_model('customers')
            >>> model.name
            'customers'

        """

        for model in self.models:
            if model.name == name:
                return model



from typing import Dict
from pydantic import BaseModel

from artefacts.mixins import ContextReader, Artifact
from artefacts.nodes import BaseNodeReference, BaseNode
from artefacts.shared import Metadata


class CatalogNode(BaseNode, BaseModel):
    pass


class CatalogNodeReference(BaseNodeReference):
    pass


class Catalog(Artifact, ContextReader, BaseModel):
    """
    Catalogs artifacts generated by `dbt docs generate`, and contain information about
    the actual state of all the tables tracked by dbt models.

    Examples:

        >>> from artefacts import Project
        >>> project = Project(target='./target')
        >>> print(project.catalog)
        <Catalog v1>

    """

    class Config:
        arbitrary_types_allowed = True

    metadata: Metadata
    nodes: Dict[CatalogNodeReference, CatalogNode]



import os
import json

from artefacts import config


class Context(object):
    def __init__(self, config=config):
        self.config = config

    def get(self, *args, **kwargs):
        return self.config.context_data.get(*args, **kwargs)

    def update(self, **kwargs):
        self.config.context_data.update(**kwargs)

    def load(self, artifact, target, allow_missing=False):
        artifact_path = os.path.join(target, artifact + ".json")

        if allow_missing and not os.path.exists(artifact_path):
            return
        else:
            with open(artifact_path, "r") as fh:
                self.config.context_data[f"raw_{artifact}"] = json.load(fh)

    def parse(self, artifact, parser, allow_missing=False):
        if allow_missing and f"raw_{artifact}" not in self.config.context_data:
            return
        else:
            self.config.context_data[f"parsed_{artifact}"] = parser(
                **self.config.context_data[f"raw_{artifact}"]
            )

    @property
    def manifest(self):
        try:
            return self.config.context_data["parsed_manifest"]
        except KeyError as err:
            raise KeyError("No parsed manifest available").with_traceback(
                err.__traceback__
            )

    @property
    def catalog(self):
        try:
            return self.config.context_data["parsed_catalog"]
        except KeyError as err:
            raise KeyError("No parsed catalog available").with_traceback(
                err.__traceback__
            )

    @property
    def run_results(self):
        try:
            return self.config.context_data["parsed_run_results"]
        except KeyError as err:
            raise KeyError("No parsed run_results available").with_traceback(
                err.__traceback__
            )

    @property
    def sources(self):
        try:
            return self.config.context_data["parsed_sources"]
        except KeyError as err:
            raise KeyError("No parsed sources available").with_traceback(
                err.__traceback__
            )



from typing import Dict, List
from pydantic import BaseModel

from artefacts.mixins import ContextReader, Artifact
from artefacts.nodes import BaseNodeReference, BaseNode
from artefacts.shared import Metadata


# https://schemas.getdbt.com/dbt/manifest/v3/index.html#nodes
class ManifestNode(BaseNode, BaseModel):
    class Config:
        extra = "allow"

    resource_type: str


class ManifestNodeReference(BaseNodeReference):
    pass


class Manifest(Artifact, ContextReader, BaseModel):
    """
    Manifests are artifacts generated by any dbt command that compiles the dbt project,
    including `dbt compile`, `dbt build`, `dbt docs generate`, etc. The manifest
    contains information about all nodes in the project, including models, tests,
    seeds, etc.

    Examples:

        >>> from artefacts import Project
        >>> project = Project(target='./target')
        >>> print(project.manifest)
        <Manifest v3>
    """

    class Config:
        arbitrary_types_allowed = True

    metadata: Metadata
    nodes: Dict[ManifestNodeReference, ManifestNode]
    child_map: Dict[ManifestNodeReference, List[ManifestNodeReference]]
    parent_map: Dict[ManifestNodeReference, List[ManifestNodeReference]]



from artefacts import config
from artefacts.context import Context


context = Context(config=config)


class ContextReader:
    @property
    def context(self):
        return context

    @property
    def manifest(self):
        return self.context.manifest

    @property
    def catalog(self):
        return self.context.catalog

    @property
    def run_results(self):
        return self.context.run_results

    @property
    def sources(self):
        return self.context.sources


class Artifact:
    @property
    def schema_version(self):
        return self.metadata.schema_version

    def __str__(self):
        return f"<{self.__class__.__name__} {self.schema_version}>"



import os
from setuptools import setup, find_packages


def get_version():
    project_dir = os.path.abspath(os.path.dirname(__file__))
    version_filename = os.path.join(project_dir, 'artefacts', 'VERSION')
    version = open(version_filename).read().strip()
    return version


setup(
    name='artefacts',
    version=get_version(),
    description='Utilities for building on dbt artifacts',
    long_description="Please see the documentation site "
                     "https://github.com/tjwaterman99/artefacts",
    author='Tom Waterman',
    license='Apache License 2.0',
    author_email='tjwaterman99@gmail.com',
    url='https://github.com/tjwaterman99/artefacts',
    packages=find_packages(),
    install_requires=[
        'pydantic==1.8.2',
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",

        'License :: OSI Approved :: Apache Software License',

        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',

        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',        
    ]
)


from typing import Union
from pydantic import BaseModel


class Metadata(BaseModel):
    class Config:
        extra = "allow"

    dbt_schema_version: str
    dbt_version: str
    user_id: Union[str, None]
    project_id: Union[str, None]
    invocation_id: Union[str, None]
    env: dict
    generated_at: str

    @property
    def schema_version(self):
        return self.dbt_schema_version.split("/")[-1].replace(".json", "")
